"""Contains application variables."""

import sqlalchemy as sa


class KeyField:
    """Represent base class for key fields."""

    def __call__(self, model):
        self.model = model
        return self

    def __repr__(self):
        return self.name

    @property
    def name(self):
        value = ''
        for i, elem in enumerate(self.__class__.__name__):
            if elem.isupper() and i > 0:
                value += '_'
            value += elem.lower()
        return value

    @property
    def label(self):
        return f'pd_{self.name}'

    @property
    def column(self):
        return sa.literal(self.value).label(self.label)

    def associated(self, obj):
        """Check if the key field associated with the given instance."""
        class_name = obj.__class__.__name__
        if self.name.startswith(class_name.lower()):
            return True
        elif self.name.startswith('run') and class_name == 'Job':
            return True
        elif self.name.startswith('process') and class_name == 'Job':
            return True
        else:
            return False


class RunId(KeyField):
    """Represents Run ID as a key field."""

    @property
    def value(self):
        if self.model.job:
            return self.model.job.record_id


class ProcessId(RunId):
    """Represents Process ID as a key field."""

    pass


class TaskId(KeyField):
    """Represents Task ID as a key field."""

    @property
    def value(self):
        if self.model.pipeline:
            return self.model.pipeline.task.id


class StepId(KeyField):
    """Represents Step ID as a key field."""

    @property
    def value(self):
        if self:
            raise NotImplementedError


run_id = ProcessId()
task_id = TaskId()
step_id = StepId()
process_id = ProcessId()



"""Contains Python API including prototypes and built-in models."""

import configparser
import git
import os
import shutil
import sys
import time as tm
import datetime as dt

import sqlalchemy as sa

from .db import db
from .config import config
from .logger import logger

from .utils import locate
from .utils import terminator
from .utils import to_process, to_python
from .wrap import check_repo


class Driver():
    """Represents application Python API."""

    def __init__(self, root=None):
        self.root = locate() if root is None else os.path.abspath(root)
        self.jobs = os.path.join(self.root, 'jobs')

        logger.configure(status=False, file=False, console=False,
                         format='[{rectype}] {message}\n',
                         alarming=False)

    def create_scheduler(self, name=None, desc=None, path=None):
        """Deploy scheduler with all necessary elements."""
        logger.debug('Creating scheduler...')
        root = self.root if path is None else os.path.abspath(path)
        samples = os.path.join(os.path.dirname(__file__), 'samples')
        dir = self.jobs
        if os.path.exists(dir) is False:
            os.makedirs(dir)
            logger.debug(f'Folder {dir} created')
        else:
            raise Exception(f'folder {dir} already exists!')
        filename = 'scheduler'
        src = os.path.join(samples, f'{filename}.txt')
        dest = os.path.join(root, f'{filename}.py')
        if os.path.exists(dest) is False:
            content = open(src, 'r').read()
            with open(dest, 'w') as fh:
                fh.write(content)
                logger.debug(f'File {dest} created')
        else:
            raise Exception(f'file {dest} already exists!')
        config_parser = configparser.ConfigParser()
        config_path = os.path.join(root, 'pydin.ini')
        config_dict = {
            'SCHEDULER': {
                'scheduler_name': name or '',
                'scheduler_desc': desc or '',
                'chargers_number': '5',
                'executors_number': '20',
                'refresh_interval': '300',
                'rerun_delay': '14400',
                'rerun_interval': '60',
                'rerun_enabled': 'True',
                'wakeup_interval': '60',
                'wakeup_enabled': 'True'
            },
            'LOGGING': {
                'console': 'True',
                'file': 'True',
                'info': 'True',
                'debug': 'False',
                'error': 'True',
                'warning': 'True',
                'critical': 'True',
                'alarming': 'False',
                'maxsize': '10485760',
                'maxdays': '1'}
            }
        config_parser.read_dict(config_dict)
        with open(config_path, 'w') as fh:
            config_parser.write(fh, space_around_delimiters=False)
        logger.debug(f'Configuration file {config_path} saved')

    def start_scheduler(self, path=None):
        """Start scheduler."""
        logger.debug('Starting scheduler...')
        root = self.root if path is None else os.path.abspath(path)
        path = os.path.join(root, 'scheduler.py')
        if os.path.exists(path) is True:
            proc = to_python(path, '--start')
            result = proc.poll()
            running = True if result is None else False
            if result is None:
                timer = 0.25
                while timer > 0:
                    wait = 0.01
                    tm.sleep(wait)
                    timer -= 0.01
                    if result != proc.poll():
                        running = False
                        break
            if running is True:
                logger.debug('Scheduler started')
                return proc.pid
            else:
                logger.debug('Scheduler failed to start')
                return None
        else:
            logger.debug('Scheduler cannot be started')
            raise Exception(f'file {path} does not exist')

    def stop_scheduler(self, path=None):
        """Stop scheduler."""
        logger.debug('Stopping scheduler...')
        path = self.root if path is None else os.path.abspath(path)
        path = os.path.join(path, 'scheduler.py')
        if os.path.exists(path) is True:
            to_python(path, 'stop')
            logger.debug('Scheduler stopped')
        else:
            logger.debug('Scheduler cannot be stopped')
            raise Exception(f'file {path} does not exist')

    def restart_scheduler(self, path=None):
        """Restart scheduler."""
        self.stop_scheduler(path=path)
        self.start_scheduler(path=path)

    def report_scheduler(self):
        """Define current scheduler status."""
        logger.debug('Checking scheduler current status...')
        conn = db.connect()
        table = db.tables.components
        select = table.select().where(table.c.id == 'SCHEDULER')
        result = conn.execute(select).first()
        status = True if result.status == 'Y' else False
        if status is False:
            logger.debug('Scheduler is not running according to status')
            return None
        else:
            pid = result.pid
            try:
                os.kill(pid, 0)
            except (OSError, ProcessLookupError):
                logger.debug('Scheduler should be running but process is dead')
                return None
            else:
                logger.debug('Scheduler is running')
                return result.pid

    def create_job(self, name=None, desc=None, mday=None,
                   hour=None, min=None, sec=None, wday=None, yday=None,
                   trig=None, start_date=None, end_date=None,
                   env=None, args=None, timeout=None, parallelism=None,
                   rerun_interval=None, rerun_limit=None, rerun_days=None,
                   rerun_period=None, sleep_period=None, wake_up_period=None,
                   alarm=None, email_list=None, debug=None, norepo=False):
        """Create job with all necessary elements."""
        logger.debug('Creating job...')
        conn = db.connect()
        table = db.tables.schedule
        values = db.normalize(job_name=name,
                              job_description=desc,
                              status=False,
                              monthday=mday,
                              hour=hour, minute=min, second=sec,
                              weekday=wday,
                              yearday=yday,
                              trigger_id=trig,
                              start_date=start_date,
                              end_date=end_date,
                              environment=env,
                              arguments=args,
                              timeout=timeout,
                              parallelism=parallelism,
                              rerun_interval=rerun_interval,
                              rerun_limit=rerun_limit,
                              rerun_days=rerun_days,
                              rerun_period=rerun_period,
                              sleep_period=sleep_period,
                              wake_up_period=wake_up_period,
                              alarm=alarm,
                              email_list=email_list,
                              debug=debug)
        logger.debug(f'Configuring schedule record with {values=}')
        insert = table.insert().values(**values)
        id = conn.execute(insert).inserted_primary_key[0]
        repr = f'Job[{id}]'
        logger.debug(f'{repr} schedule record configured')

        root = self.root
        samples = os.path.join(os.path.dirname(__file__), 'samples')
        folder = os.path.join(root, f'jobs/{id}')

        if os.path.exists(folder) is False:
            os.makedirs(folder)
            logger.debug(f'Folder {folder} created')
        else:
            raise Exception(f'folder {folder} already exists!')

        filenames = ['job', 'script']
        for filename in filenames:
            src = os.path.join(samples, f'{filename}.txt')
            dest = os.path.join(folder, f'{filename}.py')
            if os.path.exists(dest) is False:
                content = open(src, 'r').read()
                with open(dest, 'w') as fh:
                    fh.write(content)
                    logger.debug(f'File {dest} created')
            else:
                raise Exception(f'file {dest} already exists!')

        config_parser = configparser.ConfigParser()
        config_path = os.path.join(folder, 'pydin.ini')
        config_dict = {
            'JOB': {},
            'LOGGING': {
                'console': 'True',
                'file': 'True',
                'info': 'True',
                'debug': 'False',
                'error': 'True',
                'warning': 'True',
                'critical': 'True',
                'alarming': 'True',
                'maxsize': '10485760',
                'maxdays': '1'
            },
            'EMAIL': {
                'toggle': 'False'
            }
        }
        config_parser.read_dict(config_dict)
        with open(config_path, 'w') as fh:
            config_parser.write(fh, space_around_delimiters=False)
        logger.debug(f'Configuration file {config_path} saved')

        path = os.path.join(root, 'jobs/.git')
        if norepo is False and os.path.exists(path) is True:
            self.push_repo(id=id)

        logger.debug(f'{repr} successfully created and scheduled')
        return id

    def configure_job(self, id, name=None, desc=None, mday=None,
                      hour=None, min=None, sec=None, wday=None, yday=None,
                      trig=None, trig_list=None, start_date=None, end_date=None,
                      env=None, args=None, timeout=None, parallelism=None,
                      rerun_interval=None, rerun_limit=None, rerun_days=None,
                      rerun_period=None, sleep_period=None, wake_up_period=None,
                      alarm=None, email_list=None, debug=None):
        """Modify job configuration."""
        repr = f'Job[{id}]'
        logger.debug(f'Editing {repr}...')
        conn = db.connect()
        table = db.tables.schedule
        values = db.normalize(job_name=name,
                              job_description=desc,
                              monthday=mday,
                              hour=hour, minute=min, second=sec,
                              weekday=wday,
                              yearday=yday,
                              trigger_id=trig,
                              trigger_list=trig_list,
                              start_date=start_date,
                              end_date=end_date,
                              environment=env,
                              arguments=args,
                              timeout=timeout,
                              parallelism=parallelism,
                              rerun_interval=rerun_interval,
                              rerun_limit=rerun_limit,
                              rerun_days=rerun_days,
                              rerun_period=rerun_period,
                              sleep_period=sleep_period,
                              wake_up_period=wake_up_period,
                              alarm=alarm,
                              email_list=email_list,
                              debug=debug)
        if len(values) > 0:
            logger.debug(f'Configuring schedule record with {values=}')
            update = table.update().values(**values).where(table.c.id == id)
            conn.execute(update)
            logger.debug(f'{repr} edited')
        else:
            logger.debug(f'{repr} nothing to edit')

    def enable_job(self, id):
        """Change job status to Y."""
        repr = f'Job[{id}]'
        logger.debug(f'{repr} will be enabled')
        conn = db.connect()
        table = db.tables.schedule
        update = table.update().values(status='Y').where(table.c.id == id)
        result = conn.execute(update)
        if result.rowcount > 0:
            logger.debug(f'{repr} enabled')
            return True
        else:
            logger.debug(f'{repr} does not exist')
            return False

    def disable_job(self, id):
        """Change job status to N."""
        repr = f'Job[{id}]'
        logger.debug(f'{repr} will be disabled')
        conn = db.connect()
        table = db.tables.schedule
        update = table.update().values(status='N').where(table.c.id == id)
        result = conn.execute(update)
        if result.rowcount > 0:
            logger.debug(f'{repr} disabled')
            return True
        else:
            logger.debug(f'{repr} does not exist')
            return False

    def delete_job(self, id):
        """Delete particular job."""
        repr = f'Job[{id}]'
        logger.debug(f'Requested to delete {repr}')
        root = self.root
        conn = db.connect()
        table = db.tables.schedule
        delete = table.delete().where(table.c.id == id)
        conn.execute(delete)
        logger.debug(f'{repr} deleted from schedule')
        folder = os.path.join(root, f'jobs/{id}')
        shutil.rmtree(folder)
        path = os.path.join(root, 'jobs/.git')
        if os.path.exists(path) is True:
            self.push_repo(id=id)
        logger.debug(f'{repr} folder {folder} removed')

    def list_jobs(self, id=None):
        """Generate list with all scheduled jobs."""
        logger.debug('Requested to list jobs')
        conn = db.connect()
        table = db.tables.schedule
        select = table.select()
        if id is not None:
            select = select.where(table.c.id == id)
        select = select.order_by(table.c.id)
        result = conn.execute(select).fetchall()
        for row in result:
            yield dict(row)
        logger.debug('Jobs listed')

    def run_job(self, id, tag=None, date=None, record_id=None, trigger_id=None,
                recycle=None, debug=None, noalarm=None, solo=None, wait=True):
        """Run particular job."""
        repr = f'Job[{id}]'
        logger.debug(f'Requested to run {repr}')
        conn = db.connect()
        table = db.tables.schedule
        select = table.select().where(table.c.id == id)
        result = conn.execute(select).first()
        env = result.environment or 'python'
        args = result.arguments or ''
        exe = config['ENVIRONMENTS'].get(env)
        path = os.path.join(self.root, f'jobs/{id}/job.py')
        args += ' run'
        args_dict = {'tag': tag,
                     'date': date.isoformat() if date is not None else None,
                     'record': record_id,
                     'trigger': trigger_id,
                     'recycle': '' if recycle is True else None,
                     'debug': '' if debug is True else None,
                     'noalarm': '' if noalarm is False else None,
                     'solo': '' if solo is True else None}
        for key, value in args_dict.items():
            if value is not None:
                args += f' --{key} {value}'
        logger.debug(f'{exe=}, {path=}, {args=}')
        proc = to_process(exe, path, args, piped=wait)
        logger.debug(f'{repr} runs on PID {proc.pid}')
        if wait is True:
            logger.debug(f'Waiting for {repr} to finish...')
            proc.wait()
            if proc.returncode > 0:
                logger.debug(f'{repr} completed with error')
            else:
                logger.debug(f'{repr} completed')
        return proc

    def cancel_job(self, id):
        """Cancel particular job."""
        repr = f'Job[{id}]'
        logger.debug(f'Requested to cancel {repr} runs...')
        conn = db.connect()
        table = db.tables.run_history
        select = (table.select().
                  where(sa.and_(table.c.job_id == id,
                                table.c.status == 'R')))
        result = conn.execute(select).fetchall()
        for row in result:
            run = row.id
            logger.debug(f'Found {repr} running as Run[{run}]')
            self.cancel_run(run)
        logger.debug(f'All {repr} runs canceled')

    def cancel_jobs(self):
        """Cancel all jobs."""
        logger.debug('Requested to cancel all jobs...')
        conn = db.connect()
        table = db.tables.run_history
        select = table.select().where(table.c.status == 'R')
        result = conn.execute(select).fetchall()
        for row in result:
            job = row.job_id
            run = row.id
            logger.debug(f'Found Job[{job}] running as Run[{run}]')
            self.cancel_run(run)
        logger.debug('All jobs canceled')

    def cancel_run(self, id):
        """Cancel particular run."""
        repr = f'Run[{id}]'
        logger.debug(f'Requested to cancel {repr}...')
        conn = db.connect()
        table = db.tables.run_history
        select = table.select().where(table.c.id == id)
        result = conn.execute(select).first()
        if result.status == 'R':
            pid = result.pid
            try:
                terminator(pid)
            except (OSError, ProcessLookupError):
                logger.warning(f'{repr} on PID[{pid}] not found')
                update = (table.update().values(status='C').
                          where(sa.and_(table.c.id == id,
                                        table.c.status == 'R')))
                conn.execute(update)
                logger.debug(f'{repr} status changed to C')
            except Exception:
                logger.error()
            else:
                logger.debug(f'{repr} on PID[{pid}] successfully terminated')
            logger.debug(f'{repr} canceled')
        else:
            raise Exception(f'{repr} is not running')

    def deactivate_run(self, id):
        """Deactivate particular run."""
        repr = f'Run[{id}]'
        logger.debug(f'Requested to deactivate {repr}...')
        conn = db.connect()
        table = db.tables.run_history
        select = table.select().where(table.c.id == id)
        result = conn.execute(select).first()
        deactivated = dt.datetime.now()
        if result.status in ('W', 'E', 'T'):
            update = (table.update().values(deactivated=deactivated).
                      where(sa.and_(table.c.id == id,
                                    table.c.status.in_(['W', 'E', 'T']))))
            conn.execute(update)
            logger.debug(f'{repr} deactivated')
        else:
            raise Exception(f'{repr} cannot be deactivated')

    def create_config(self):
        """Create global config."""
        logger.debug('Creating global config...')
        from .config import user_config
        dirpath = os.path.dirname(user_config)
        if not os.path.exists(dirpath):
            os.mkdir(dirpath)
        if not os.path.exists(user_config):
            config_parser = configparser.ConfigParser()
            config_dict = {
                'GENERAL': {
                    'debug': '',
                    'editor': '',
                    'owner': ''
                },
                'DATABASE': {
                    'vendor_name': '',
                    'driver_name': '',
                    'client_path': '',
                    'path': '',
                    'host': '',
                    'port': '',
                    'sid': '',
                    'service_name': '',
                    'username': '',
                    'password': ''
                },
                'EMAIL': {
                    'toggle': '',
                    'host': '',
                    'port': '',
                    'tls': '',
                    'address': '',
                    'username': '',
                    'password': ''
                },
                'API': {
                    'host': '',
                    'port': '',
                    'token': ''
                },
                'ENVIRONMENTS': {
                    'python': sys.executable
                }
            }
            config_parser.read_dict(config_dict)
            with open(user_config, 'w') as fh:
                config_parser.write(fh, space_around_delimiters=False)
            logger.debug(f'Global config {user_config} created')
            return user_config
        else:
            raise Exception(f'file {user_config} already exists!')

    def create_repo(self, url=None):
        """Create git repository."""
        logger.debug('Creating git repo...')
        try:
            repo = git.Repo(self.jobs)
        except git.exc.InvalidGitRepositoryError:
            repo = git.Repo.init(self.jobs)
            logger.debug(f'Git repo {repo.common_dir} initiated')
            repo.create_remote('origin', url)
            logger.debug('Remote <origin> created')
            commit = repo.index.commit('Initial commit')
            logger.debug(f'Initial commit made as <{commit.hexsha}>')
            origin = repo.remote('origin')
            origin.push(repo.head.ref.name, set_upstream=True)
            logger.debug(f'Git repo {repo.common_dir} published in {url}')

            filename = 'gitignore'
            samples = os.path.join(os.path.dirname(__file__), 'samples')
            src = os.path.join(samples, f'{filename}.txt')
            dest = os.path.join(self.jobs, '.gitignore')
            if os.path.exists(dest) is False:
                content = open(src, 'r').read()
                with open(dest, 'w') as fh:
                    fh.write(content)
                    logger.debug(f'File {dest} created')
        else:
            raise Exception(f'git repo in {repo.common_dir} already exists!')
        return repo

    @check_repo
    def push_repo(self, id=None, message=None):
        """Commit all current changes and push git repo to the remote."""
        logger.debug('Pushing git repo...')
        timestamp = int(tm.time())
        repo = git.Repo(self.jobs)
        if id is None:
            repo.git.add(all=True)
            logger.debug('All files staged')
        else:
            repr = f'Job[{id}]'
            logger.debug(f'Requested to push only {repr}')
            folder = os.path.join(self.jobs, str(id))
            untracked_files = repo.untracked_files
            unstaged_files = [diff.a_path for diff in repo.index.diff(None)]
            all_files = [*untracked_files, *unstaged_files]
            for file in all_files:
                abspath = os.path.join(self.jobs, file)
                commonpath = os.path.commonpath([folder, abspath])
                if folder == commonpath:
                    logger.debug(f'Found {repr} file {file} for stage')
                    repo.index.add(file)

        staged_files = repo.index.diff(repo.head.name)
        count = len(staged_files)
        logger.debug(f'{count} files staged')
        if count > 0:
            message = message or f'pydin[{timestamp}]'
            commit = repo.index.commit(message)
            logger.debug(f'This {message} commit made as <{commit.hexsha}>')
            origin = repo.remote('origin')
            result = origin.push(repo.head.ref.name)
            logger.debug(f'Git repo {repo.common_dir} successfully pushed')
            return result
        else:
            logger.debug(f'Nothing to push in {repo.common_dir}')

    @check_repo
    def pull_repo(self):
        """Pull git repo from the remote to get all external changes."""
        logger.debug('Pulling git repo...')
        repo = git.Repo(self.jobs)
        origin = repo.remote('origin')
        result = origin.pull()
        logger.debug(f'Git repo {repo.common_dir} successfully pulled')
        return result



"""Contains main application configurator and other configuration elements."""

import os
import sys
import configparser


home = os.environ.get('PYDIN_HOME')
filename = 'pydin.ini'
user_config = os.path.abspath(os.path.expanduser(f'~/.pydin/{filename}'))
home_config = os.path.join(home, filename) if home is not None else None
local_config = os.path.join(os.path.dirname(sys.argv[0]), filename)


class Configurator(dict):
    """Represents main application configurator."""

    def __init__(self):
        super().__init__()
        tables = ['GENERAL', 'DATABASE', 'SMTP', 'API',
                  'ENVIRONMENTS', 'SCHEDULER', 'JOB',
                  'LOGGING', 'EMAIL']
        for table in tables:
            self[table] = self.Table(name=table)

    class Table(dict):
        """Represents configuration table."""

        def __init__(self, name, **kwargs):
            self.name = name
            super().__init__(**kwargs)

        def __setitem__(self, key, value):
            """Normalize given item key and set item value."""
            super().__setitem__(key.lower(), value)

        def __getitem__(self, key):
            """Normalize given item key and get item value if it exists."""
            return super().__getitem__(key.lower())

        def get(self, key):
            """Normalize given item key and get item value or return None."""
            return super().get(key.lower())

        def has_record(self, record):
            """Check if configuration table record exists."""
            if self.get(record) is not None:
                return True
            else:
                return False

        pass

    def load(self, paths, encoding=None):
        """Load extetrnal configuration from files to configurator."""
        paths = [path for path in paths if path is not None]
        parser = configparser.ConfigParser(allow_no_value=True)
        parser.read(paths, encoding=encoding)
        for section in parser.sections():
            for option in parser.options(section):
                value = parser[section][option]
                if (
                    value is None
                    or value.upper() == 'NONE'
                    or value.isspace() is True
                ):
                    self[section][option] = None
                elif value.upper() == 'TRUE':
                    self[section][option] = True
                elif value.upper() == 'FALSE':
                    self[section][option] = False
                elif value.isdigit() is True:
                    self[section][option] = int(value)
                elif value.isdecimal() is True:
                    self[section][option] = float(value)
                else:
                    self[section][option] = value
        return self

    def save(self):
        """Save current settings to the user configuration file."""
        config_parser = configparser.ConfigParser()
        config_dict = {'GENERAL': self['GENERAL'],
                       'DATABASE': self['DATABASE'],
                       'EMAIL': self['EMAIL'],
                       'API': self['API'],
                       'ENVIRONMENTS': self['ENVIRONMENTS']}
        config_parser.read_dict(config_dict)
        with open(user_config, 'w', encoding='utf8') as fh:
            config_parser.write(fh, space_around_delimiters=False)

    def has_table(self, table):
        """Check if configuration table exists."""
        if self.get(table) is not None:
            return True
        else:
            return False


config = Configurator()
config.load([user_config, home_config, local_config])


###############################################################################
#                              Default parameters
###############################################################################
DEBUG = False

LOG_CONSOLE = True
LOG_FILE = True
LOG_INFO = True
LOG_DEBUG = False
LOG_ERROR = True
LOG_WARNING = True
LOG_CRITICAL = True
LOG_ALARMING = False
LOG_MAXSIZE = 10485760
LOG_MAXDAYS = 1

DATABASE_VENDOR_NAME = None
DATABASE_DRIVER_NAME = None
DATABASE_PATH = None
DATABASE_HOST = None
DATABASE_PORT = None
DATABASE_SID = None
DATABASE_SERVICE_NAME = None
DATABASE_USERNAME = None
DATABASE_PASSWORD = None

EMAIL_TOGGLE = False
EMAIL_HOST = None
EMAIL_PORT = None
EMAIL_TLS = False
EMAIL_ADDRESS = None
EMAIL_USERNAME = None
EMAIL_PASSWORD = None

API_HOST = None
API_PORT = None
API_TOKEN = None

EDITOR = None
OWNER = None

SCHEDULER_NAME = None
SCHEDULER_DESC = None
SCHEDULER_CHARGERS_NUMBER = 5
SCHEDULER_EXECUTORS_NUMBER = 20
SCHEDULER_REFRESH_INTERVAL = 300
SCHEDULER_RERUN_DELAY = 14400
SCHEDULER_RERUN_ENABLED = True
SCHEDULER_RERUN_INTERVAL = 60
SCHEDULER_WAKEUP_INTERVAL = 60
SCHEDULER_WAKEUP_ENABLED = True

ENV_PYTHON = sys.executable



"""Contains web API server."""

import datetime as dt
import json
import os
import platform
import signal
import subprocess as sp
import sys

import flask
import flask_httpauth

from .api import Driver

from .db import db
from .config import config
from .logger import logger

from .utils import locate
from .utils import to_process
from .utils import to_boolean


app = flask.Flask(__name__)
auth = flask_httpauth.HTTPTokenAuth(scheme='Bearer')
root = os.environ['PYDIN_HOME'] if os.environ.get('PYDIN_HOME') else locate()


OK = flask.Response(status=200)
BAD_REQUEST = flask.Response(status=400)
SERVER_ERROR = flask.Response(status=500)


class Server():
    """Represents web API server."""

    def __init__(self, host=None, port=None, dev=False):
        self.app = app
        argv = sys.argv.copy()
        dev = True if 'dev' in argv else dev

        self.host = host or '0.0.0.0'
        self.port = port or 8080
        if config['API'].get('host'):
            self.host = config['API'].get('host')
        if config['API'].get('host'):
            self.port = config['API'].get('port')

        self.server_name = platform.node()
        self.user_name = os.getlogin()
        self.start_date = None
        self.stop_date = None
        self.status = None
        self.debug = self.dev = dev if isinstance(dev, bool) else False

        logger.configure(file=False, format='[{rectype}] {message}\n',
                         alarming=False)

        if len(argv) > 1:
            if argv[1] == 'start':
                self.start()
            elif argv[1] == 'stop':
                self.stop()

    @property
    def url(self):
        """Get server url."""
        return f'http://{self.host}:{self.port}'

    def start(self):
        """Start API server."""
        conn = db.connect()
        table = db.tables.components
        select = table.select().where(table.c.id == 'RESTAPI')
        result = conn.execute(select).first()
        status = to_boolean(result.status)
        if not status:
            self.status = True
            self.start_date = dt.datetime.now()
            app = 'pydin.web:app'
            env = os.environ.copy()
            env['PYDIN_HOME'] = locate()
            if self.dev:
                script = ['flask', 'run']
                args = ['--host', self.host, '--port', str(self.port)]
                cmd = [arg for arg in [*script, *args] if arg is not None]
                mode = 'development'
                env['FLASK_APP'] = app
                env['FLASK_ENV'] = mode
                try:
                    self.proc = sp.Popen(cmd, env=env)
                    self.log()
                    self.proc.wait()
                except KeyboardInterrupt:
                    self.status = False
                    self.stop_date = dt.datetime.now()
                    self.log()
                    self.proc.terminate()
            else:
                script = 'waitress-serve'
                args = ['--host', self.host, '--port', str(self.port), app]
                args = [arg for arg in args if arg is not None]
                self.proc = to_process(script, args=args, env=env)
                self.log()
        else:
            pid = result.pid
            print(f'Server already working on PID[{pid}]')

    def stop(self):
        """Stop API server."""
        self.status = False
        self.stop_date = dt.datetime.now()
        conn = db.connect()
        table = db.tables.components
        select = table.select().where(table.c.id == 'RESTAPI')
        result = conn.execute(select).first()
        status = to_boolean(result.status)
        pid = result.pid
        if status and pid:
            try:
                os.kill(pid, signal.SIGTERM)
            except Exception as error:
                print(f'Server on PID[{pid}] stopped with error: {error}')
        update = table.update().where(table.c.id == 'RESTAPI')
        update = update.values(status='N', stop_date=self.stop_date)
        conn.execute(update)

    def log(self):
        conn = db.connect()
        table = db.tables.components
        update = table.update().where(table.c.id == 'RESTAPI')
        update = update.values(status='Y' if self.status else 'N',
                               user_name=self.user_name,
                               server_name=self.server_name,
                               pid=self.proc.pid,
                               debug='Y' if self.debug else None,
                               url=self.url,
                               start_date=self.start_date,
                               stop_date=self.stop_date)
        conn.execute(update)

    pass


@auth.verify_token
def verify_token(token):
    """."""
    TOKEN = config['API']['token']
    if token == TOKEN:
        return token


@app.route('/api/test')
@auth.login_required
def test():
    """Test server working capacity with a simple request."""
    return 'Hello there!'


@app.route('/api/help')
@auth.login_required
def help():
    """."""
    demo = os.path.join(os.path.dirname(__file__), 'demo/web.txt')
    lines = open(demo, 'r').readlines()
    text = ''.join(lines)
    return text


@app.route('/api/start-scheduler', methods=['POST'])
@auth.login_required
def start_scheduler():
    """."""
    driver = Driver(root=root)
    pid = driver.start_scheduler()
    if pid is None:
        return SERVER_ERROR
    else:
        return OK


@app.route('/api/stop-scheduler', methods=['POST'])
@auth.login_required
def stop_scheduler():
    """."""
    driver = Driver(root=root)
    driver.stop_scheduler()
    return OK


@app.route('/api/create-job', methods=['POST'])
@auth.login_required
def create_job():
    """."""
    driver = Driver(root=root)
    driver.create_job()
    return OK


@app.route('/api/enable-job', methods=['POST'])
@auth.login_required
def enable_job():
    """."""
    request = flask.request
    id = request.args.get('id')
    if id is not None and id.isdigit():
        driver = Driver(root=root)
        result = driver.enable_job(id)
        if result is True:
            return OK
        else:
            return SERVER_ERROR
    else:
        return BAD_REQUEST


@app.route('/api/disable-job', methods=['POST'])
@auth.login_required
def disable_job():
    """."""
    request = flask.request
    id = request.args.get('id')
    if id is not None and id.isdigit():
        driver = Driver(root=root)
        result = driver.disable_job(id)
        if result is True:
            return OK
        else:
            return SERVER_ERROR
    else:
        return BAD_REQUEST


@app.route('/api/run-job', methods=['POST'])
@auth.login_required
def run_job():
    """."""
    request = flask.request
    id = request.args.get('id')
    if id is not None and id.isdigit():
        tag = request.args.get('tag')
        date = request.args.get('date')
        record_id = request.args.get('process')
        trigger_id = request.args.get('trigger')
        recycle = json.loads(request.args.get('recycle', 'false'))
        debug = json.loads(request.args.get('debug', 'false'))
        solo = json.loads(request.args.get('solo', 'false'))
        wait = json.loads(request.args.get('wait', 'false'))
        kwargs = {}
        if tag:
            kwargs['tag'] = int(tag)
        elif date:
            kwargs['date'] = dt.datetime.fromisoformat(date)
        elif record_id:
            kwargs['record_id'] = int(record_id)
        elif trigger_id:
            kwargs['trigger_id'] = int(trigger_id)
        kwargs['recycle'] = recycle
        kwargs['debug'] = debug
        kwargs['solo'] = solo
        driver = Driver(root=root)
        proc = driver.run_job(id, wait=wait, **kwargs)
        if wait:
            proc.wait()
            if proc.returncode > 0:
                return SERVER_ERROR
            else:
                return OK
        else:
            return OK
    else:
        return BAD_REQUEST


@app.route('/api/cancel-run', methods=['POST'])
@auth.login_required
def cancel_run():
    """."""
    request = flask.request
    process_id = request.args.get('id')
    if process_id and process_id.isdigit():
        process_id = int(process_id)
        driver = Driver(root=root)
        try:
            driver.cancel_run(process_id)
        except Exception:
            return BAD_REQUEST
        else:
            return OK


@app.route('/api/deactivate-run', methods=['POST'])
@auth.login_required
def deactivate_run():
    """."""
    request = flask.request
    process_id = request.args.get('id')
    if process_id and process_id.isdigit():
        process_id = int(process_id)
        driver = Driver(root=root)
        try:
            driver.deactivate_run(process_id)
        except Exception:
            return BAD_REQUEST
        else:
            return OK


@app.route('/api/sync-repo', methods=['POST'])
@auth.login_required
def sync_repo():
    """."""
    request = flask.request
    id = request.args.get('id')
    driver = Driver(root=root)
    driver.pull_repo()
    kwargs = {}
    if id is not None and id.isdigit():
        kwargs['id'] = int(id)
    driver.push_repo(**kwargs)
    return OK



"""Contains data source interfaces."""

import platform
import importlib as imp
import ftplib

import pepperoni as pe

import paramiko as pm
# import sqlalchemy as sa


class Localhost():
    """Represents localhost i.e. no connection required."""

    def __init__(self):
        self.host = platform.node()


class Server():
    """Represents ordinary remote server."""

    def __init__(self, name=None, host=None, port=None, user=None,
                 password=None, key=None, keyfile=None,
                 ssh=False, sftp=False, ftp=False):
        self.name = name.lower() if isinstance(name, str) else None
        self.host = host if isinstance(host, str) else None
        self.port = int(port) if isinstance(port, (str, int)) else None

        creds = {}
        creds['user'] = user if isinstance(user, str) else None
        creds['password'] = password if isinstance(password, str) else None
        creds['key'] = key if isinstance(key, str) else None
        creds['keyfile'] = keyfile if isinstance(keyfile, str) else None
        self.creds = creds

        self.ssh = ssh if isinstance(ssh, bool) else False
        self.sftp = sftp if isinstance(sftp, bool) else False
        self.ftp = ftp if isinstance(ftp, bool) else False

        cache = imp.import_module('pydin.cache')
        cache.CONNECTIONS[f'{self.name}'] = self

    def __repr__(self):
        """Represent server as its name."""
        return self.name

    class Connection():
        """Represents ordinary remote server connection."""

        def __init__(self, server, creds):
            server = server if isinstance(server, Server) else None
            user = creds.get('user')
            password = creds.get('password')
            keyfile = creds.get('keyfile')
            if server.ssh is True and server is not None:
                self.ssh = pm.SSHClient()
                self.ssh.set_missing_host_key_policy(pm.AutoAddPolicy())
                self.ssh.connect(server.host, port=server.port,
                                 username=user, password=password,
                                 key_filename=keyfile)
            if server.sftp is True and server is not None:
                if server.ssh is True:
                    self.sftp = self.ssh.open_sftp()
                else:
                    transport = pm.Transport((server.host, server.port))
                    if keyfile is not None:
                        key = pm.RSAKey(filename=keyfile)
                        transport.connect(username=user, pkey=key)
                    else:
                        transport.connect(username=user,
                                          password=password)
                    self.sftp = pm.SFTPClient.from_transport(transport)
            if server.ftp is True and server is not None:
                self.ftp = ftplib.FTP(host=server.host,
                                      user=user,
                                      passwd=password)

        def execute(self, command, **kwargs):
            """Execute given command in server terminal."""
            result = self.ssh.exec_command(command, **kwargs)
            stdout = result[1].read().decode()
            stderr = result[2].read().decode()
            return stdout, stderr

    @property
    def connection(self):
        """Get active server connection."""
        if hasattr(self, '_connection'):
            return self._connection
        return None

    @connection.setter
    def connection(self, value):
        if isinstance(value, self.Connection):
            self._connection = value

    def connect(self):
        """Connect to remote server."""
        if not self.connection:
            self.connection = self.Connection(self, self.creds)
        return self.connection

    def exists(self, path):
        """Check if file with given path exists in file system."""
        conn = self.connect()
        if conn:
            if self.sftp:
                try:
                    conn.sftp.chdir(path)
                    return True
                except IOError:
                    return False
            elif self.ftp:
                try:
                    conn.ftp.cwd(path)
                    return True
                except Exception:
                    return False
        else:
            raise ConnectionError(f'not connected to {self.name}')


class Database(pe.Database):
    """Represents database server."""

    def __init__(self, name=None, vendor_name=None, driver_name=None,
                 path=None, host=None, port=None, sid=None, service_name=None,
                 user=None, password=None):
        self.name = name.lower() if isinstance(name, str) else None
        super().__init__(vendor=vendor_name, driver=driver_name,
                         path=path, host=host, port=port,
                         sid=sid, service=service_name,
                         user=user, password=password)

        cache = imp.import_module('pydin.cache')
        cache.CONNECTIONS[f'{self.name}'] = self

    def __repr__(self):
        """Represent database as its name."""
        return self.name



"""Utilits."""

import os
import sys
import signal
import ctypes

import re
import datetime as dt
import calendar as cnd

import abc
import configparser

import threading as th
import subprocess as sp
import importlib as imp

import pepperoni as pe

import sqlalchemy as sa
import sqlalchemy.dialects as sad
import sqlparse as spa

from .sources import Server, Database
from .const import LINUX, MACOS, WINDOWS
from .wrap import classproperty, staticproperty


def locate():
    """Get current location and move to it.

    Returns
    -------
    root : str
        Path to current location as a simple string.
    """
    root = os.path.abspath(os.path.dirname(sys.argv[0]))
    os.chdir(root)
    return root


def declare(object):
    """Declare some object as a module level attribute.

    Returns
    -------
    value : bool
        Always True value meaning that object was successfully declared.
    """
    namespace = imp.import_module('pydin')
    name = object.__class__.__name__.lower()
    setattr(namespace, name, object)
    return True


def cache(object):
    """Save object in application cache.

    Parameters
    ----------
    object : any
        Object that must be registered in memory.

    Returns
    -------
    value : bool
        Always True value indicating that object was successfully stored.
    """
    namespace = imp.import_module('pydin.cache')
    name = object.__class__.__name__.lower()
    setattr(namespace, name, object)
    return True


def importer(module_name, *args):
    """Get importing object from the given module."""
    namespace = imp.import_module(module_name)
    degree = len(args)
    for i, attribute_name in enumerate(args, start=1):
        attribute = getattr(namespace, attribute_name)
        if i < degree:
            namespace = attribute
        elif callable(attribute):
            return attribute()
        else:
            return attribute
    return namespace


def terminator(pid):
    """Terminate process at a given PID correctly."""
    if LINUX or MACOS:
        os.kill(pid, signal.SIGTERM)
    elif WINDOWS:
        kernel = ctypes.windll.kernel32
        kernel.FreeConsole()
        kernel.AttachConsole(pid)
        kernel.SetConsoleCtrlHandler(None, 1)
        kernel.GenerateConsoleCtrlEvent(0, 0)


def case(value, *options):
    """Parse the given arguments into a case-logic."""
    n = len(options)
    for i, option in enumerate(options):
        if i % 2 == 0 and i < n-1 and value == option:
            return options[i+1]
        elif i % 2 == 0 and i == n-1:
            return option


def coalesce(*args):
    """Pick first non-None value from the list of variables.

    Parameters
    ----------
    args
        List of varialbes to check.

    Returns
    -------
    value
        First met non-None value.
    """
    for arg in args:
        if arg is not None:
            return arg


def first(array):
    """Get the first value from the array."""
    if array:
        if isinstance(array, dict):
            key_of_first_value = list(array)[0]
            return array[key_of_first_value]
        elif isinstance(array, (list, tuple)):
            return array[0]
        elif isinstance(array, set):
            return list(array)[0]


def last(array):
    """Get the last value from the array."""
    if array:
        if isinstance(array, dict):
            key_of_last_value = list(array)[-1]
            return array[key_of_last_value]
        elif isinstance(array, (list, tuple)):
            return array[-1]
        elif isinstance(array, set):
            return list(array)[-1]


def to_integer(value):
    """."""
    if isinstance(value, int):
        return value
    elif isinstance(value, float):
        return int(value)
    elif isinstance(value, str) and value.isdigit():
        return int(value)
    elif value is None:
        return value


def to_boolean(value):
    """."""
    if isinstance(value, bool):
        return value
    elif isinstance(value, str):
        return True if value == 'Y' else False
    else:
        return None


def to_none(value):
    """."""
    if not isinstance(value, bool):
        return None


def to_timestamp(value):
    """Convert initial value to timestamp.

    Parameters
    ----------
    value : int, float, str, datetime or None
        Initial value that must be converted.
        Note that str must be in ISO format!

    Returns
    -------
    value : int
        Converted value.
    """
    if isinstance(value, (int, float)) is True:
        return int(value)
    elif isinstance(value, str) is True:
        return int(dt.datetime.fromisoformat(value).timestamp())
    elif isinstance(value, dt.datetime) is True:
        return int(value.timestamp())
    elif value is None:
        return value


def to_date(value):
    """Convert initial value to date object."""
    if isinstance(value, str) is True:
        return dt.date.fromisoformat(value)
    elif isinstance(value, dt.date) is True:
        return value
    elif value is None:
        return value


def to_datetime(value):
    """Convert initial value to datetime object."""
    if isinstance(value, (int, float)) is True:
        return dt.datetime.fromtimestamp(to_timestamp(value))
    elif isinstance(value, str) is True:
        return dt.datetime.fromisoformat(value).replace(microsecond=0)
    elif isinstance(value, dt.datetime) is True:
        return value.replace(microsecond=0)
    elif value is None:
        return value


def to_thread(name, function):
    """Initialize and return a new thread using given parameters."""
    thread = th.Thread(name=name, target=function, daemon=True)
    return thread


def to_process(exe, path=None, args=None, env=None, piped=False):
    """Run the command in a separate process."""
    command = []
    if re.match(r'^.*(\\|/).*$', exe):
        exe = os.path.abspath(exe)
    command.append(exe)
    if path:
        path = os.path.abspath(path)
        command.append(path)
    if args:
        if isinstance(args, str):
            args = args.split()
        command.extend(args)
    kwargs = {}
    kwargs['env'] = env
    kwargs['stdin'] = sp.PIPE if piped else sp.DEVNULL
    kwargs['stdout'] = sp.PIPE if piped else sp.DEVNULL
    kwargs['stderr'] = sp.PIPE if piped else sp.DEVNULL
    if WINDOWS:
        kwargs['creationflags'] = sp.CREATE_NO_WINDOW
    proc = sp.Popen(command, **kwargs)
    return proc


def to_python(path, args=None):
    """Run the Python script in a separate process."""
    python = sys.executable
    proc = to_process(python, path, args)
    return proc


def to_lower(value):
    """Get the string in lowercase."""
    if isinstance(value, str):
        return value.lower()
    else:
        return value


def to_upper(value):
    """Get the string in uppercase."""
    if isinstance(value, str):
        return value.upper()
    else:
        return value


def camel_to_human(s):
    """Conver CamelCase style into ordinary human text."""
    return ''.join([' '+c if c.isupper() else c for c in s]).lstrip()


def to_sql(text):
    """Format given SQL text."""
    result = spa.format(text, keyword_case='upper',
                        identifier_case='lower',
                        reindent_aligned=True)
    return result


def is_path(string):
    """Check whether given string is a path or not."""
    if string is None:
        return False
    elif os.path.exists(string):
        return True
    elif string.startswith('/'):
        return True
    elif string[1:].startswith(':\\'):
        return True
    elif '/' in string:
        return True
    elif '\\' in string:
        return True
    else:
        return False


def read_file_or_string(value):
    """Read content from file or string."""
    if os.path.isfile(value):
        return open(value, 'r').read()
    else:
        return value


def sql_formatter(query, expand_select=None, expand_where=None,
                  make_subquery=False, make_empty=False,
                  fix_new_lines=False, db=None):
    """Modify the given SQL query."""
    import sqlparse.sql as s
    import sqlparse.tokens as t

    def is_select(token):
        if isinstance(token, s.IdentifierList):
            return True
        return False

    def is_where(token):
        if isinstance(token, s.Where):
            return True
        return False

    def has_where(statement):
        for token in statement:
            if is_where(token):
                return True
        return False

    def get_length(statement):
        if statement.tokens:
            return len(statement.tokens)
        return 0

    def as_subquery(statement, db):
        if db.vendor == 'postgresql':
            return f'select * from (\n{statement}) as s\n'
        else:
            return f'select * from (\n{statement})\n'

    parsed = spa.parse(query)
    statement = parsed[0] if parsed else None

    ws = s.Token(t.Whitespace, ' ')
    nl = s.Token(t.Newline, '\n')
    cm = s.Token(t.Punctuation, ',')
    where = s.Token(t.Keyword, 'where')
    and_ = s.Token(t.Operator, 'and')

    if expand_select:
        expansion = s.Token(t.Other, expand_select)
        for token in statement.tokens:
            if is_select(token):
                tokens = s.TokenList([cm, nl, expansion])
                last_position = get_length(token)
                token.insert_after(last_position, tokens)

    if expand_where:
        expansion = s.Token(t.Other, expand_where)
        if has_where(statement):
            tokens = s.TokenList([ws, expansion, nl, and_, ws])
            for token in statement.tokens:
                if is_where(token):
                    token.insert_after(1, tokens)
        else:
            tokens = s.TokenList([nl, where, ws, expansion, nl])
            last_position = get_length(statement)
            statement.insert_after(last_position, tokens)

    if make_subquery:
        statement = as_subquery(statement, db)
    if make_empty:
        select_from = as_subquery(statement, db)
        statement = f'{select_from} where 1 = 0'

    result = spa.format(str(statement), keyword_case='upper',
                        identifier_case='lower', reindent_aligned=True)
    string = str(result)
    if fix_new_lines:
        if re.match(r'INSERT.+SELECT', string):
            string = string.replace('SELECT', '\nSELECT')
    return string


def sql_preparer(obj, db):
    """Prepare the given object for compilation."""
    if isinstance(obj, dt.datetime):
        if db.vendor == 'oracle':
            string = f'{obj:%Y-%m-%d %H:%M:%S}'
            fmt = 'YYYY-MM-DD HH24:MI:SS'
            return sa.func.to_date(string, fmt)
        elif db.vendor == 'sqlite':
            return f'{obj:%Y-%m-%d %H:%M:%S}'
        else:
            return f'{obj:%Y-%m-%d %H:%M:%S}'
    else:
        return obj


def sql_compiler(obj, db):
    """Compile the given object into SQL expression."""
    try:
        database = getattr(sad, db.vendor)
        if database:
            dialect = database.dialect()
            result = obj.compile(dialect=dialect, compile_kwargs=db.ckwargs)
            return str(result)
        else:
            raise NotImplementedError
    except Exception:
        result = obj.compile(compile_kwargs=db.ckwargs)
        return str(result)


def sql_converter(value, db):
    """Convert the given value into SQL expression."""
    if isinstance(value, int):
        return value
    elif isinstance(value, str):
        return f'\'{value}\''
    elif isinstance(value, dt.datetime):
        if db.vendor == 'oracle':
            string = f'{value:%Y-%m-%d %H:%M:%S}'
            fmt = 'YYYY-MM-DD HH24:MI:SS'
            return f'to_date(\'{string}\', \'{fmt}\')'
        else:
            return f'\'{value:%Y-%m-%d %H:%M:%S}\''
    else:
        return value


def get_version():
    """Get current application version."""
    return importer('pydin', '__version__')


def get_job():
    """Get active application job."""
    return importer('pydin', 'job')


def get_logger():
    """Get application core logger."""
    return importer('pydin', 'logger')


def get_email():
    """Get application email interface."""
    return importer('pydin', 'logger', 'email')


def get_credentials():
    """Get a special data dictionary containing saved credentials."""
    return importer('pepperoni', 'credentials')


def get_system_information():
    """Get a special data dictionary containing system information."""
    return importer('pepperoni', 'sysinfo')


def installed():
    """Check if application is installed and configured."""
    from .config import user_config, config
    if not os.path.exists(user_config):
        return False
    elif not config.get('DATABASE'):
        return False
    elif not config['DATABASE'].get('vendor_name'):
        return False
    elif not config.get('API'):
        return False
    elif not config.get('ENVIRONMENTS'):
        return False
    else:
        return True


class Logging():
    """Represents logging configurator used for special loggers."""

    def __init__(self, task_logging=True, step_logging=True,
                 query_logging=False, file_logging=False,
                 database=None, schema_name=None):
        self.metadata = sa.MetaData()
        attributes = ['task', 'step', 'query', 'file']
        for attribute in attributes:
            arg_name = f'{attribute}_logging'
            arg_value = locals()[arg_name]

            logging_name = attribute.lower()
            logging_type = attribute.capitalize()
            logging_status = True if arg_value else False
            table_name = arg_value if isinstance(arg_value, str) else None
            config = arg_value if isinstance(arg_value, dict) else None
            logging = self.make_logging(logging_type, logging_status,
                                        database=database,
                                        schema_name=schema_name,
                                        table_name=table_name,
                                        config=config)
            setattr(self, logging_name, logging)

    def make_logging(self, logging_type, logging_status, database=None,
                     schema_name=None, table_name=None, config=None):
        """Configure one of the logging items."""
        module = sys.modules[__name__]
        class_name = f'{logging_type}Log'
        constructor = getattr(module, class_name)
        if logging_status:
            table_name = coalesce(table_name, constructor.default_name)
            if config:
                table_name = config.get('table_name', table_name)
                schema_name = config.get('schema_name', schema_name)
                database = config.get('database', database)
        else:
            schema_name, table_name = None, None
            database = None
        return constructor(self, database=database,
                           schema_name=schema_name,
                           table_name=table_name)


class BaseLog(abc.ABC):
    """Represents base interface for table logging."""

    def __init__(self, logging, database=None, schema_name=None,
                 table_name=None):
        if isinstance(logging, Logging):
            self.logging = logging
        else:
            class_name = logging.__class__.__name__
            message = f'logging must be Logging, not {class_name}'
            raise TypeError(message)
        if isinstance(database, Database):
            self.database = database
        elif isinstance(database, str):
            self.database = connector.receive(database)
        else:
            module = imp.import_module('pydin.db')
            self.database = module.db
        schema_name = schema_name if isinstance(schema_name, str) else None
        table_name = table_name if isinstance(table_name, str) else None
        self.schema_name, self.table_name = schema_name, table_name

    def __str__(self):
        class_name = camel_to_human(self.__class__.__name__)
        return f'{class_name}ging [{self.table_name}]'

    def __repr__(self):
        class_name = self.__class__.__name__
        return f'{class_name}ging({self.table_name})'

    @abc.abstractmethod
    def default_name(self):
        """Get the default table name used for this logging."""

    @property
    @abc.abstractmethod
    def columns(self):
        """Get list of columns used in logging."""

    @abc.abstractmethod
    def create_table(self):
        """Create database table used in logging."""
        schema_name = self.schema_name
        table_name = self.table_name
        columns = self.columns
        metadata = self.logging.metadata
        table = sa.Table(table_name, metadata, *columns, schema=schema_name)
        table.create(self.database.engine, checkfirst=True)
        return table

    @abc.abstractmethod
    def setup(self, name, date_column=None):
        """Configure logger."""
        if not hasattr(self, 'logger'):
            logger = pe.logger(name, table=True, console=False, file=False,
                               email=False)
            if self.table_name is None:
                logger.configure(table=False)
            else:
                if not self.database.engine.has_table(self.table_name):
                    self.create_table()
                options = {'database': self.database, 'name': self.table_name}
                if date_column:
                    options['date_column'] = date_column
                logger.configure(db=options)
            self.logger = logger
        return self


class TaskLog(BaseLog):
    """Represents task logging interface."""

    @staticproperty
    def default_name():
        return 'pd_task_history'

    @property
    def columns(self):
        seq_name = f'{self.table_name}_seq'
        columns = [sa.Column('id', sa.Integer, sa.Sequence(seq_name),
                             primary_key=True,
                             comment='Unique task ID'),
                   sa.Column('job_id', sa.Integer,
                             comment='Job that performed this task'),
                   sa.Column('run_id', sa.Integer,
                             comment='Run that performed this task'),
                   sa.Column('task_name', sa.String(30),
                             comment='Name of the task'),
                   sa.Column('task_date', sa.DateTime,
                             comment='Date for which task was performed'),
                   sa.Column('start_date', sa.DateTime,
                             comment='Date when task started'),
                   sa.Column('end_date', sa.DateTime,
                             comment='Date when task ended'),
                   sa.Column('status', sa.String(1),
                             comment='Current status.'),
                   sa.Column('records_read', sa.Integer,
                             comment='Total number of read records'),
                   sa.Column('records_written', sa.Integer,
                             comment='Total number of written records'),
                   sa.Column('records_error', sa.Integer,
                             comment='Total number of occurred errors'),
                   sa.Column('result_value', sa.Integer,
                             comment='Short numeric execution result'),
                   sa.Column('result_long', sa.String(512),
                             comment='Long string execution result'),
                   sa.Column('updated', sa.DateTime,
                             comment='Date of this record change')]
        return columns

    def create_table(self):
        return super().create_table()

    def setup(self):
        logger_name = 'pydin.task.logger'
        return super().setup(logger_name, date_column='updated')


class StepLog(BaseLog):
    """Represents step logging interface."""

    @staticproperty
    def default_name():
        return 'pd_step_history'

    @property
    def columns(self):
        seq_name = f'{self.table_name}_seq'
        columns = [sa.Column('id', sa.Integer, sa.Sequence(seq_name),
                             primary_key=True,
                             comment='Unique step ID'),
                   sa.Column('job_id', sa.Integer,
                             comment='Job that performed this step'),
                   sa.Column('run_id', sa.Integer,
                             comment='Run that performed this step'),
                   sa.Column('task_id', sa.Integer,
                             comment='Task ID of this step'),
                   sa.Column('step_name', sa.String(30),
                             comment='Name of the step'),
                   sa.Column('step_type', sa.String(3),
                             comment='Type of the step'),
                   sa.Column('step_a', sa.String(30),
                             comment='Name of the step item A'),
                   sa.Column('step_b', sa.String(30),
                             comment='Name of the step item b'),
                   sa.Column('step_c', sa.String(30),
                             comment='Name of the step item c'),
                   sa.Column('step_date', sa.DateTime,
                             comment='Date for which step was performed'),
                   sa.Column('start_date', sa.DateTime,
                             comment='Date when task started'),
                   sa.Column('end_date', sa.DateTime,
                             comment='Date when task ended'),
                   sa.Column('status', sa.String(1),
                             comment='Current status.'),
                   sa.Column('records_read', sa.Integer,
                             comment='Total number of read records'),
                   sa.Column('records_written', sa.Integer,
                             comment='Total number of written records'),
                   sa.Column('records_error', sa.Integer,
                             comment='Total number of occurred errors'),
                   sa.Column('result_value', sa.Integer,
                             comment='Short numeric result of execution'),
                   sa.Column('result_long', sa.String(512),
                             comment='Long string result of execution'),
                   sa.Column('updated', sa.DateTime,
                             comment='Date of this record change')]
        return columns

    def create_table(self):
        return super().create_table()

    def setup(self, step):
        logger_name = f'pydin.step.{step.seqno}.logger'
        return super().setup(logger_name, date_column='updated')


class QueryLog(BaseLog):
    """Represents SQL logger configurator."""

    @staticproperty
    def default_name():
        return 'pd_query_log'

    @property
    def columns(self):
        seq_name = f'{self.table_name}_seq'
        columns = [sa.Column('id', sa.Integer, sa.Sequence(seq_name),
                             primary_key=True,
                             comment='Unique file ID'),
                   sa.Column('job_id', sa.Integer,
                             comment='Job that processed the file'),
                   sa.Column('run_id', sa.Integer,
                             comment='Run that processed the file'),
                   sa.Column('task_id', sa.Integer,
                             comment='Task that processed the file'),
                   sa.Column('step_id', sa.Integer,
                             comment='Step that processed the file'),
                   sa.Column('db_name', sa.String(25),
                             comment='Database of SQL execution'),
                   sa.Column('table_name', sa.String(128),
                             comment='Name of involved table'),
                   sa.Column('query_type', sa.String(10),
                             comment='Type of SQL being executed'),
                   sa.Column('query_text', sa.String(10),
                             comment='Text of SQL being executed'),
                   sa.Column('start_date', sa.DateTime,
                             comment='Date when SQL execution started'),
                   sa.Column('end_date', sa.DateTime,
                             comment='Date when SQL execution ended'),
                   sa.Column('output_rows', sa.Integer,
                             comment='Total number of returned rows'),
                   sa.Column('output_text', sa.Integer,
                             comment='Returned database message'),
                   sa.Column('error_code', sa.Integer,
                             comment='Occurred database error code'),
                   sa.Column('error_text', sa.Integer,
                             comment='Occurred database error text')]
        return columns

    def create_table(self):
        return super().create_table()

    def setup(self):
        logger_name = 'pydin.query.logger'
        return super().setup(logger_name)

    def start(self, model):
        """Start query logging."""
        job_id = model.job.id if model.job else None
        run_id = model.job.record.id if model.job else None
        task_id = model.task.id
        step_id = model.step.id
        db_name = model.db.name
        table_name = model.table_name
        query_type = model.query_type or model.model_type
        query_text = model.query_text
        self.logger.root.table.new()
        self.logger.table(job_id=job_id,
                          run_id=run_id,
                          task_id=task_id,
                          step_id=step_id,
                          db_name=db_name,
                          table_name=table_name,
                          query_type=query_type,
                          query_text=query_text,
                          start_date=dt.datetime.now())

    def end(self, model):
        """End query logging."""
        output_rows = model.output_rows
        output_text = model.output_text
        error_text = model.error_text
        error_code = model.error_code
        if not error_code and error_text:
            error_pattern = None
            if model.db.vendor == 'oracle':
                error_pattern = r'\w{3}-\d{3,5}'
            if error_pattern and re.match(error_pattern, error_text):
                error_code = re.search(error_pattern, error_text).group()
        self.logger.table(end_date=dt.datetime.now(),
                          output_rows=output_rows,
                          output_text=output_text,
                          error_code=error_code,
                          error_text=error_text)


class FileLog(BaseLog):
    """Represents file logger configurator."""

    @staticproperty
    def default_name():
        return 'pd_file_log'

    @property
    def columns(self):
        seq_name = f'{self.table_name}_seq'
        columns = [sa.Column('id', sa.Integer, sa.Sequence(seq_name),
                             primary_key=True,
                             comment='Unique file ID'),
                   sa.Column('job_id', sa.Integer,
                             comment='Job that processed the file'),
                   sa.Column('run_id', sa.Integer,
                             comment='Run that processed the file'),
                   sa.Column('task_id', sa.Integer,
                             comment='Task that processed the file'),
                   sa.Column('step_id', sa.Integer,
                             comment='Step that processed the file'),
                   sa.Column('server', sa.String(30),
                             comment='Host name having the file'),
                   sa.Column('file_name', sa.String(256),
                             comment='Name of the file'),
                   sa.Column('file_date', sa.DateTime,
                             comment='File modification time'),
                   sa.Column('file_size', sa.Integer,
                             comment='File size in bytes'),
                   sa.Column('start_date', sa.DateTime,
                             comment='Date when file process started'),
                   sa.Column('end_date', sa.DateTime,
                             comment='Date when file process ended')]
        return columns

    def create_table(self):
        return super().create_table()

    def setup(self):
        logger_name = 'pydin.file.logger'
        return super().setup(logger_name)

    def start(self, model, fileinfo):
        """Start file logging."""
        self.logger.root.table.new()
        job_id = model.job.id if model.job else None
        run_id = model.job.record.id if model.job else None
        task_id = model.task.id
        step_id = model.step.id
        server_name = fileinfo['server'].host
        file_name = fileinfo['file']
        file_date = fileinfo['mtime']
        file_size = fileinfo['size']
        self.logger.table(job_id=job_id,
                          run_id=run_id,
                          task_id=task_id,
                          step_id=step_id,
                          server_name=server_name,
                          file_name=file_name,
                          file_date=file_date,
                          file_size=file_size,
                          start_date=dt.datetime.now())

    def end(self, model):
        """End file logging."""
        self.logger.table(end_date=dt.datetime.now())


class Calendar():
    """Represents calendar configurator."""

    class Date():
        """Represents a single date."""

        def __init__(self, now=None):
            self.now = now
            self.timezone = None

        def __repr__(self):
            """Represent object as a datetime string."""
            return str(self.now)

        def __getattribute__(self, name):
            if '(' in name and name.endswith(')'):
                func_name = name[:name.index('(')]
                if hasattr(self, func_name):
                    expression = f'self.{name}'
                    attribute = eval(expression)
                    return attribute
                else:
                    return super().__getattribute__(name)
            else:
                return super().__getattribute__(name)

        class value():
            """Represents wrapper used to calculate final date value."""

            def __init__(self, func):
                self.func = func
                self.__doc__ = func.__doc__

            def __call__(self, obj):
                """Get final date value."""
                datetime = self.func(obj)
                if obj.timezone is not None:
                    return datetime.astimezone(tz=obj.timezone)
                else:
                    return datetime

        @property
        @value
        def now(self):
            """Get current date."""
            return self._now

        @now.setter
        def now(self, value):
            if isinstance(value, dt.datetime) or value is None:
                self._now = value.replace(microsecond=0)

        @property
        @value
        def start(self):
            """Get start of current date."""
            return self._start

        @property
        @value
        def end(self):
            """Get end of current date."""
            return self._end

        @property
        def timezone(self):
            return self._timezone

        @timezone.setter
        def timezone(self, value):
            if value is None or value is dt.timezone.utc:
                self._timezone = value
            elif isinstance(value, int):
                timezone = dt.timezone(dt.timedelta(hours=value))
                self._timezone = timezone
            elif isinstance(value, str):
                if value.upper() == 'UTC':
                    self._timezone = dt.timezone.utc
                elif value.isdigit():
                    hours = int(value)
                    self._timezone = dt.timezone(dt.timedelta(hours=hours))
            elif isinstance(value, (list, tuple)):
                hours, minutes = value[0], value[1]
                timedelta = dt.timedelta(hours=hours, minutes=minutes)
                self._timezone = dt.timezone(timedelta)

        @property
        def prev(self):
            """Get previous day."""
            now = self.now-dt.timedelta(days=1)
            return Calendar.Date(now)

        @property
        def next(self):
            """Get next day."""
            now = self.now+dt.timedelta(days=1)
            return Calendar.Date(now)

        @property
        def hour(self):
            """Get current hour."""
            return Calendar.Hour(self.now)

        @property
        def yesterday(self):
            """Get yesterday."""
            return Calendar.Yesterday(self.now)

        @property
        def tomorrow(self):
            """Get tomorrow."""
            return Calendar.Tomorrow(self.now)

        @property
        def month(self):
            """Get current month."""
            return Calendar.Month(self.now)

        @property
        def year(self):
            """Get current year."""
            return Calendar.Year(self.now)

        @property
        def local(self):
            """Set local timezone."""
            self.timezone = None
            return self

        @property
        def utc(self):
            """Set UTC timezone."""
            self.timezone = dt.timezone.utc
            return self

        @property
        def yd(self):
            """Shortcut for yesterday."""
            return self.yesterday

        @property
        def tm(self):
            """Shortcut for tomorrow."""
            return self.tomorrow

        @property
        def y(self):
            """Shortcut for year."""
            return self.year

        @property
        def dd(self):
            """Shortcut for day."""
            return self

        @property
        def mm(self):
            """Shortcut for month."""
            return self.month

        @property
        def hh(self):
            """Shortcut for hour."""
            return self.hour

        @property
        def pv(self):
            """Shrotcut for prev."""
            return self.prev

        @property
        def nt(self):
            """Shrotcut for next."""
            return self.next

        def days_back(self, delta):
            delta = dt.timedelta(days=delta)
            now = self._now-delta
            return Calendar.Day(now)

        def hours_back(self, delta):
            delta = dt.timedelta(hours=delta)
            now = self._now-delta
            return Calendar.Hour(now)

        def minutes_back(self, delta):
            delta = dt.timedelta(minutes=delta)
            now = self._now-delta
            return Calendar.Day(now)

        def seconds_back(self, delta):
            delta = dt.timedelta(seconds=delta)
            now = self._now-delta
            return Calendar.Day(now)

        def months_back(self, delta):
            now = self.now
            while delta > 0:
                delta -= 1
                now = now.replace(day=1)-dt.timedelta(days=1)
            return Calendar.Month(now)

        def week_back(self, delta):
            pass

        def years_back(self, delta):
            pass

        def minutes_round(self, level):
            date = self._now
            micro = date.microsecond
            seconds = date.second
            minutes = date.minute % level
            delta = dt.timedelta(seconds=seconds, minutes=minutes,
                                 microseconds=micro)
            now = date-delta
            return Calendar.Date(now)

    class Day(Date):
        """Represents certain day."""

        def __init__(self, now):
            super().__init__(now)

        @property
        def _start(self):
            """Get start of current date."""
            return self._now.replace(hour=0, minute=0, second=0)

        @property
        def _end(self):
            """Get end of current date."""
            return self._now.replace(hour=23, minute=59, second=59)

    class Today(Day):
        """Represents current day."""

        def __init__(self):
            super().__init__(dt.datetime.now())

    class Yesterday(Day):
        """Represents yesterday."""

        def __init__(self, now=None):
            super().__init__((now or dt.datetime.now())-dt.timedelta(days=1))

    class Tomorrow(Day):
        """Represents tomorrow."""

        def __init__(self, now=None):
            super().__init__((now or dt.datetime.now())+dt.timedelta(days=1))

    class Hour(Date):
        """Represents certain hour."""

        @property
        def _start(self):
            """Get hour start."""
            return self._now.replace(minute=0, second=0)

        @property
        def _end(self):
            """Get hour end."""
            return self._now.replace(minute=59, second=59)

        @property
        def prev(self):
            """Get previous hour."""
            now = self._now-dt.timedelta(hours=1)
            return Calendar.Hour(now)

        @property
        def next(self):
            """Get next hour."""
            now = self._now+dt.timedelta(hours=1)
            return Calendar.Hour(now)

    class Month(Date):
        """Represents certain month."""

        @property
        def _start(self):
            """Get month start."""
            return self._now.replace(day=1, hour=0, minute=0, second=0)

        @property
        def _end(self):
            """Get month end."""
            day = cnd.monthrange(self._now.year, self._now.month)[1]
            return self._now.replace(day=day, hour=23, minute=59, second=59)

        @property
        def prev(self):
            """Get previous month."""
            now = self._now.replace(day=1)-dt.timedelta(days=1)
            return Calendar.Month(now)

    class Week(Date):
        """Represents certain week."""

    class Year(Date):
        """Represents certain year."""

        @property
        def _start(self):
            """Get year start."""
            return self._now.replace(month=1, day=1, hour=0,
                                     minute=0, second=0)

        @property
        def _end(self):
            """Get year end."""
            return self._now.replace(month=12, day=31, hour=23,
                                     minute=59, second=59)

        @property
        def prev(self):
            """Get previous year."""
            now = self._now-dt.timedelta(days=365)
            return Calendar.Year(now)


class Connector(dict):
    """Represents connections configurator."""

    def __init__(self):
        super().__init__()
        self.config = os.path.expanduser('~/.pydin/sources.ini')

    class Connection():
        """Represents connection objects factory."""

        def __new__(cls, name=None, **options):
            """Create connection object using options passed."""
            given_options = options.keys()
            database_options = {'vendor_name', 'driver_name', 'database'}
            server_options = {'protocol', 'host', 'port'}

            if set.intersection(database_options, given_options):
                vendor_name = options['vendor_name']
                driver_name = options.get('driver_name')
                database = options.get('database')
                username = options.get('username')
                password = options.get('password')
                host = options.get('host')
                port = options.get('port')
                path = None
                if is_path(database):
                    path = os.path.abspath(database)
                    database = None

                sid = options.get('sid')
                tnsname = options.get('tnsname')
                database = coalesce(database, sid, tnsname)

                service = options.get('service')
                service_name = options.get('service_name')
                service_name = coalesce(service, service_name)

                return Database(name=name, vendor_name=vendor_name,
                                driver_name=driver_name,
                                user=username, password=password,
                                path=path, host=host, port=port,
                                sid=database, service_name=service_name)
            elif set.intersection(server_options, given_options):
                protocol = options.get('protocol').lower()
                host = options.get('host')
                port = options.get('port')
                username = options.get('username')
                password = options.get('password')
                key = options.get('key')
                keyfile = options.get('keyfile')

                ssh = True if protocol == 'ssh' else False
                sftp = True if protocol == 'sftp' else False
                ftp = True if protocol == 'ftp' else False

                return Server(name=name, host=host, port=port,
                              user=username, password=password,
                              key=key, keyfile=keyfile,
                              ssh=ssh, sftp=sftp, ftp=ftp)
            else:
                return None

    def load(self, path=None, encoding=None):
        """Load configuration from file."""
        path = path or self.config
        parser = configparser.ConfigParser(allow_no_value=True)
        parser.read(path, encoding=encoding)
        for section in parser.sections():
            connection_name = section.lower()
            self[connection_name] = {}
            for option in parser.options(section):
                parameter_name = option.lower()
                parameter_value = parser[section][option]
                self[connection_name][parameter_name] = parameter_value
        return self

    def receive(self, name):
        """Receive connection object using its name."""
        if isinstance(name, str):
            name = name.lower()
            try:
                cache = imp.import_module('pydin.cache')
                return cache.CONNECTIONS[name]
            except KeyError:
                try:
                    options = self[name]
                except KeyError as error:
                    path = self.config
                    message = f'no record named <{name}> found, check {path}'
                    traceback = error.__traceback__
                    raise Exception(message).with_traceback(traceback)
                else:
                    return self.Connection(name=name, **options)
        else:
            class_name = name.__class__.__name__
            message = f'name must be str, not {class_name}'
            raise TypeError(message)


calendar = Calendar()
connector = Connector()
connector.load()



"""Contains main application database schema and objects."""

import os
import datetime as dt

import pepperoni as pe
import sqlalchemy as sa
import cx_Oracle as oracle

from .config import config
from .utils import installed
from .utils import to_datetime


class Database(pe.Database):
    """Represents database schema and objects."""

    def __init__(self):
        if config['DATABASE'].get('vendor_name'):
            self.configure()
        self.null = self.Null()
        pass

    def __repr__(self):
        """Represent this database as path, SID or service name."""
        if self.path is not None:
            return f'Database[{self.path}]'
        elif self.sid is not None:
            return f'Database[{self.sid}]'
        elif self.service is not None:
            return f'Database[{self.service}]'

    class Tables():
        """Represents database tables."""

        def __init__(self, database):
            self.database = database
            self.schedule = self.load('pd_schedule')
            self.pipeline_config = self.load('pd_pipeline_config')
            self.node_config = self.load('pd_node_config')
            self.run_history = self.load('pd_run_history')
            self.task_history = self.load('pd_task_history')
            self.step_history = self.load('pd_step_history')
            self.sql_log = self.load('pd_query_log')
            self.file_log = self.load('pd_file_log')
            self.components = self.load('pd_components')

        def load(self, table_name=None):
            if table_name:
                return self.database.table(table_name)

        pass

    class Record():
        """."""

        def __init__(self, database, table, id=None):
            self.database = database
            self.table = table
            self.id = id
            pass

        def select(self, id):
            """."""
            self.id = id
            return self.id

        def create(self):
            """."""
            conn = db.connect()
            table = self.table
            columns = [c.name for c in table.columns]
            primary_keys = [pk.name for pk in table.primary_key.columns]
            values = {c: None for c in columns if c not in primary_keys}
            insert = table.insert().values(**values)
            result = conn.execute(insert)
            self.id = result.inserted_primary_key[0]
            return self.id

        def read(self):
            """."""
            if self.id:
                conn = self.database.connect()
                table = self.table
                select = table.select().where(table.c.id == self.id)
                result = conn.execute(select).first()
                return result
            else:
                return None
            pass

        def write(self, **kwargs):
            """."""
            if self.id:
                conn = self.database.connect()
                table = self.table
                updated = dt.datetime.now()
                update = (table.update().values(updated=updated, **kwargs).
                          where(table.c.id == self.id))
                result = conn.execute(update)
                return result
            else:
                return None
            pass

        pass

    class Null(sa.sql.elements.Null):
        """Represent null data type."""

        def __repr__(self):
            """Represent as NULL."""
            return 'NULL'

        pass

    def record(self, table, id=None):
        """."""
        return self.Record(self, table, id=id)

    def configure(self):
        vendor_name = config['DATABASE']['vendor_name']
        driver_name = config['DATABASE'].get('driver_name')
        path = config['DATABASE'].get('path')
        host = config['DATABASE'].get('host')
        port = config['DATABASE'].get('port')
        sid = config['DATABASE'].get('sid')
        service_name = config['DATABASE'].get('service_name')
        username = config['DATABASE'].get('username')
        password = config['DATABASE'].get('password')
        client_path = config['DATABASE'].get('client_path')
        if client_path and vendor_name == 'oracle':
            oracle.init_oracle_client(lib_dir=client_path)
        super().__init__(vendor=vendor_name, driver=driver_name,
                         path=path, host=host, port=port,
                         sid=sid, service=service_name,
                         user=username, password=password)

    def load(self):
        """Load application database schema."""
        self.tables = self.Tables(self)

    def deploy(self):
        """Deploy application database schema."""
        if self.vendor == 'sqlite':
            if not os.path.exists(self.path):
                open(self.path, 'w')
        print('WARNING: Automatic DB schema deployment not implemented yet. '
              'Please deploy the schema yourself using scripts from GitHub.')

    def normalize(self, job_name=None, job_description=None, status=None,
                  monthday=None, hour=None, minute=None, second=None,
                  weekday=None, yearday=None,
                  trigger_id=None, trigger_list=None,
                  start_date=None, end_date=None,
                  environment=None, arguments=None,
                  timeout=None, parallelism=None,
                  rerun_interval=None, rerun_limit=None, rerun_days=None,
                  rerun_period=None, sleep_period=None, wake_up_period=None,
                  alarm=None, email_list=None, debug=None):
        """Normalize parameters in accordance with their data types."""
        values = {}
        setup = [{'column': 'job_name', 'value': job_name,
                  'types': (str,)},
                 {'column': 'job_description', 'value': job_description,
                  'types': (str,)},
                 {'column': 'status', 'value': status,
                  'types': (bool, ),
                  'norm_func': lambda arg: 'Y' if arg is True else 'N'},
                 {'column': 'monthday', 'value': monthday,
                  'types': (int, str)},
                 {'column': 'hour', 'value': hour,
                  'types': (int, str)},
                 {'column': 'minute', 'value': minute,
                  'types': (int, str)},
                 {'column': 'second', 'value': second,
                  'types': (int, str)},
                 {'column': 'weekday', 'value': weekday,
                  'types': (int, str)},
                 {'column': 'yearday', 'value': yearday,
                  'types': (int, str)},
                 {'column': 'trigger_id', 'value': trigger_id,
                  'types': (int,)},
                 {'column': 'trigger_list', 'value': trigger_list,
                  'types': (str,)},
                 {'column': 'start_date', 'value': start_date,
                  'types': (str, dt.datetime), 'norm_func': to_datetime},
                 {'column': 'end_date', 'value': end_date,
                  'types': (str, dt.datetime), 'norm_func': to_datetime},
                 {'column': 'environment', 'value': environment,
                  'types': (str,)},
                 {'column': 'arguments', 'value': arguments,
                  'types': (str,)},
                 {'column': 'timeout', 'value': timeout,
                  'types': (int,)},
                 {'column': 'parallelism', 'value': parallelism,
                  'types': (int, str)},
                 {'column': 'rerun_interval', 'value': rerun_interval,
                  'types': (int,)},
                 {'column': 'rerun_limit', 'value': rerun_limit,
                  'types': (int,)},
                 {'column': 'rerun_days', 'value': rerun_days,
                  'types': (int,)},
                 {'column': 'rerun_period', 'value': rerun_period,
                  'types': (int, str)},
                 {'column': 'sleep_period', 'value': sleep_period,
                  'types': (int, str)},
                 {'column': 'wake_up_period', 'value': wake_up_period,
                  'types': (int, str)},
                 {'column': 'alarm', 'value': alarm,
                  'types': (bool,),
                  'norm_func': lambda arg: 'Y' if arg is True else self.null},
                 {'column': 'email_list', 'value': email_list,
                  'types': (str,)},
                 {'column': 'debug', 'value': debug,
                  'types': (bool,),
                  'norm_func': lambda arg: 'Y' if arg is True else self.null}]

        for item in setup:
            column = item['column']
            value = item['value']
            types = item['types']
            norm_func = item.get('norm_func')
            if isinstance(value, types) or value is self.null:
                if value is self.null or not callable(norm_func):
                    values[column] = value
                elif callable(norm_func):
                    values[column] = norm_func(value)
        return values


db = Database()
if installed():
    db.load()



"""Python ETL framework."""

from .api import Driver
from .cli import Manager
from .web import Server
from .core import Scheduler, Job, Pipeline

from .models import Mapper
from .models import Table, SQL, Select, Insert
from .models import CSV, JSON, XML
from .models import Filenames, FileManager

from .models import Model
from .models import Extractable, Transformable, Loadable, Executable

from .logger import logger
from .utils import Logging, Calendar
from .utils import calendar, connector
from .utils import get_version, get_system_information
from .utils import get_job, get_logger, get_email, get_credentials

from .fields import run_id, task_id, step_id, process_id

__version__ = '0.2.0'
__status__ = 'Production'

__author__ = 'Timur Faradzhov'
__email__ = 'timurfaradzhov@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2023, The PyDin Project'
__maintainer__ = __author__

__all__ = [Driver, Manager, Server, Scheduler, Job, Pipeline, Mapper,
           Model, Extractable, Transformable, Loadable, Executable,
           Table, SQL, Select, Insert, CSV, JSON, XML, Filenames, FileManager,
           Logging, Calendar, logger, calendar, connector,
           get_version, get_system_information, get_logger, get_email,
           get_credentials, get_job, run_id, task_id, step_id, process_id]



"""Contains application manager."""

import argparse
import getpass
import time
import datetime as dt
import os
import re
import subprocess as sp
import sys

from .api import Driver

from .db import db
from .config import config
from .logger import logger
from .utils import connector
from .utils import locate
from .utils import to_none


class Manager():
    """Represents application manager which is a common CLI."""

    def __init__(self):
        self.root = locate()
        self.driver = Driver()

        args = self._parse_arguments()
        self.sure = args.sure
        self.wait = args.wait
        self.output = args.output
        self.error = args.error

        logger.configure(status=False, file=False, console=False,
                         format='[{rectype}] {message}\n',
                         alarming=False)

        # If argv more than one then command was entered.
        command = [arg for arg in sys.argv if arg.startswith('-') is False]
        self.parse(command)

    def parse(self, command):
        """Parse and execute command."""
        if len(command) == 1:
            self.start_console()
        elif len(command) == 2 and command[1] == 'help':
            self.help()
        elif len(command) == 2 and command[1] == 'install':
            self.install()
        elif len(command) > 2 and command[-1] == 'help':
            name = '_'.join(command[1:2 if len(command) == 3 else 3])
            self.help(name=name)
        elif len(command) > 2:
            # Read command. Command length is two words as maximum.
            try:
                name = '_'.join(command[1:3])
                args = command[3:] if len(command) > 3 else []
                func = getattr(self, name)
            except AttributeError:
                self.unknown()
            else:
                try:
                    func(*args)
                except TypeError as e:
                    if (
                        e.args[0].startswith(f'{name}() takes')
                        or e.args[0].startswith(f'{name}() missing')
                       ):
                        self.incorrect(name)
                    else:
                        logger.error()
                except Exception:
                    logger.error()
        else:
            self.unknown()

    def help(self, name=None):
        """Show help note for all available commands."""
        if name == 'help' or name is None:
            funcs = ['start_console', 'install',
                     'create_scheduler',
                     'start_scheduler', 'stop_scheduler',
                     'restart_scheduler',
                     'create_job', 'edit_script',
                     'configure_job',
                     'enable_job', 'disable_job', 'delete_job',
                     'list_jobs',
                     'run_job', 'run_jobs', 'cancel_job', 'cancel_jobs',
                     'cancel_run', 'deactivate_run',
                     'create_config', 'edit_config',
                     'create_repo', 'sync_repo']
            docs = {}
            for func in funcs:
                doc = getattr(self, func).__doc__.splitlines()[0]
                docs[func] = doc
            demo = os.path.join(os.path.dirname(__file__), 'demo/cli.txt')
            text = open(demo, 'r').read().format(**docs)
            print(text)
        elif name in dir(self):
            func = getattr(self, name)
            text = func.__doc__
            print(text)
        else:
            self.unknown()

    def incorrect(self, name):
        """Print message about incorrect command."""
        name = name.replace('_', ' ')
        print(f'Incorrect command. Type *{name} help* to see details.')

    def unknown(self):
        """Print message about unknown command."""
        print('Unknown command. Type *help* to get list of commands.')

    def start_console(self):
        """Start interactive console."""
        print('Hello there!')
        while True:
            now = dt.datetime.now()
            command = input(f'{now:%H:%M} $ ')
            if command == 'quit':
                break
            elif command == '':
                continue
            else:
                command = ['manager.py', *command.split()]
                self.parse(command)
        print('Good bye!')

    def install(self):
        """Install application files in current location."""
        import pydin as pd

        self._header()
        print('Welcome!')
        print(f'This is an installation of PyDin v.{pd.__version__}.')
        while True:
            sure = input('Continue? [Y/n] ')
            if sure in ('Y', 'n'):
                break
        if sure == 'Y':
            self.create_config()
            self.create_scheduler()
            self._configure_database()
            self._expand_configs()

    def create_scheduler(self):
        """Create scheduler in current location."""
        print('Please follow the steps to create the scheduler.')
        name = input('Enter the name or leave empty:\n')
        desc = input('Enter the description or leave empty:\n')
        self.driver.create_scheduler(name=name, desc=desc)
        print('Scheduler created in current location.')

    def start_scheduler(self):
        """Start scheduler."""
        pid = self.driver.start_scheduler()
        if pid is not None:
            print(f'Scheduler started on PID {pid}.')
        else:
            print('Scheduler failed to start')

    def stop_scheduler(self):
        """Stop scheduler."""
        self.driver.stop_scheduler()
        print('Scheduler stopped.')

    def restart_scheduler(self):
        """Restart scheduler."""
        self.stop_scheduler()
        self.start_scheduler()
        print('Scheduler restarted.')

    def report_scheduler(self):
        """Show current scheduler status."""
        pid = self.driver.report_scheduler()
        if pid is not None:
            print(f'Scheduler is running on PID {pid}.')
        else:
            print('Scheduler is not running.')

    def create_job(self, scenario=None):
        """Create job in scheduler of current location."""
        if scenario is None:
            print('Please follow the steps to create the job.')
            configuration = self._configure_job()
        sure = self._are_you_sure()
        if sure:
            print('Creating job...')
            if scenario is None:
                id = self.driver.create_job(**configuration)
            elif scenario == 'default':
                id = self.driver.create_job()
            print(f'Job created with ID {id}.')
        else:
            print('Operation canceled.')

    def configure_job(self, id):
        """Configure job data in schedule table."""
        id = self._check_id(id)
        print('Please follow the steps to configure the job.')
        configuration = self._configure_job()
        sure = self._are_you_sure()
        if sure:
            print('Editing job...')
            self.driver.configure_job(id, **configuration)
            print('Job edited.')
        else:
            print('Operation canceled.')

    def edit_script(self, id):
        """Open job script in text editor."""
        id = self._check_id(id)
        editor = config['GENERAL'].get('editor')
        path = os.path.join(self.root, f'jobs/{id}/script.py')
        if os.path.exists(path) is True:
            # Launch edition process and wait until it is completed.
            proc = sp.Popen([editor, path])
            print(f'Editing {path}...')
            proc.wait()
        else:
            print(f'File {path} does not exists.')

    def enable_job(self, id):
        """Activate job."""
        id = self._check_id(id)
        result = self.driver.enable_job(id)
        if result is True:
            print(f'Job {id} enabled.')
        else:
            print(f'Job {id} does not exist.')

    def disable_job(self, id):
        """Deactivate job."""
        id = self._check_id(id)
        result = self.driver.disable_job(id)
        if result is True:
            print(f'Job {id} disabled.')
        else:
            print(f'Job {id} does not exist.')

    def delete_job(self, id):
        """Delete job."""
        id = self._check_id(id)
        repr = f'Job[{id}]'
        print(f'You are trying to delete {repr}...')
        sure = input('Are you sure? [delete] ')
        sure = True if sure == 'delete' else False
        if sure is True:
            print(f'{repr} delete confirmed.')
            try:
                self.driver.delete_job(id)
            except Exception:
                print(f'{repr} deleted with errors.')
            else:
                print(f'{repr} deleted.')
        else:
            print(f'{repr} delete was not confirmed.')

    def list_jobs(self, id=None):
        """Show all scheduled jobs."""
        id = self._check_id(id) if id is not None else id
        jobs = list(self.driver.list_jobs(id=id))
        row = '|{id:<4}|{name:<30}|{desc:<35}|{status:<6}|'
        head = row.format(id='ID',
                          name='JOB_NAME',
                          desc='JOB_DESCRIPTION',
                          status='STATUS')
        border = f'+{"":->78}+'
        print(border)
        print(head)
        for job in jobs:
            id = job['id']
            name = job['job_name'] or ''
            if len(name) > 30:
                name = f'{name[:27]}...'
            desc = job['job_description'] or ''
            if len(desc) > 35:
                desc = f'{desc[:31]}...'
            status = job['status'] or ''
            print(border)
            print(row.format(id=id, name=name, desc=desc, status=status))
        print(border)

    def run_job(self, id, *args):
        """Run one chosen job."""
        id = self._check_id(id)
        repr = f'Job[{id}]'
        ln = len(args)
        keys = ('tag', 'date', 'process', 'trigger')
        key = args[0] if ln > 1 and args[0] in keys else None
        value = args[1] if ln > 1 and key is not None else None
        kwargs = {}
        if key is None:
            now = dt.datetime.now()
            timestamp = int(now.timestamp())
            print(f'You are trying to start {repr} for current date...')
            kwargs['tag'] = timestamp
        elif key == 'tag':
            tag = int(value)
            print(f'You are trying to start {repr} using timestamp {tag}...')
            kwargs['tag'] = tag
        elif key == 'date':
            date = dt.datetime.fromisoformat(value)
            print(f'You are trying to start {repr} using date {date}...')
            kwargs['date'] = date
        elif key == 'process':
            record_id = int(value)
            print(f'You are trying to start {repr} as Run[{record_id}]...')
            kwargs['record_id'] = record_id
        elif key == 'trigger':
            trigger_id = int(value)
            print(f'You are trying to trig {repr} from Run[{trigger_id}]...')
            kwargs['trigger_id'] = trigger_id
        kwargs['recycle'] = True if 'recycle' in args else None
        kwargs['debug'] = True if 'debug' in args else None
        kwargs['noalarm'] = False if 'noalarm' in args else None
        kwargs['solo'] = True if 'solo' in args else None

        sure = self.sure or self._are_you_sure()
        if sure:
            proc = self.driver.run_job(id, wait=False, **kwargs)
            now = dt.datetime.now()
            date = '%Y-%m-%d %H:%M:%S'
            print(f'{repr} started on PID {proc.pid}, {now:{date}}.')
            if self.wait:
                print(f'Waiting for {repr} to finish...')
                try:
                    proc.wait()
                except KeyboardInterrupt:
                    now = dt.datetime.now()
                    print(f'\n{repr} canceled, {now:{date}}.')
                else:
                    now = dt.datetime.now()
                    if proc.returncode > 0:
                        print(f'{repr} completed with error, {now:{date}}.')
                        if self.error is True:
                            print(proc.stderr.read().decode())
                    else:
                        print(f'{repr} completed, {now:{date}}.')
                        if self.output is True:
                            print(proc.stdout.read().decode())

    def run_jobs(self, path=None):
        """Run many jobs."""
        path = os.path.abspath(path or 'run.list')
        if os.path.exists(path) is True:
            for row in open(path, 'r').readlines():
                row = row.split()
                ln = len(row)
                if ln >= 1:
                    id = row[0]
                    args = row[1:] if ln > 1 else []
                    self.run_job(id, *args)
                    time.sleep(1)
            clean = input(f'Clean file {path}? [Y] ')
            if clean == 'Y':
                open(path, 'w').close()

    def cancel_job(self, id):
        """Cancel one chosen job runs."""
        id = self._check_id(id)
        repr = f'Job[{id}]'
        print(f'You are trying to cancel all {repr} runs...')
        sure = self.sure or self._are_you_sure()
        if sure:
            self.driver.cancel_job(id)
            print(f'All {repr} runs canceled.')

    def cancel_jobs(self):
        """Cancel all currently running jobs."""
        print('You are trying to cancel all currently running jobs...')
        sure = self.sure or self._are_you_sure()
        if sure:
            self.driver.cancel_jobs()
            print('All jobs canceled.')

    def cancel_run(self, id):
        """Cancel run using its ID."""
        id = self._check_id(id)
        repr = f'Run[{id}]'
        print(f'You are trying to cancel {repr}...')
        sure = self.sure or self._are_you_sure()
        if sure:
            self.driver.cancel_run(id)
            print(f'{repr} canceled.')

    def deactivate_run(self, id):
        """Deactivate run using its ID."""
        id = self._check_id(id)
        repr = f'Run[{id}]'
        print(f'You are trying to deactivate {repr}...')
        sure = self.sure or self._are_you_sure()
        if sure:
            self.driver.deactivate_run(id)
            print(f'{repr} deactivated.')

    def create_config(self):
        """Create global configuration file."""
        config_path = self.driver.create_config()
        if config_path is not None:
            print(f'Global config created ({config_path}).')
            return config.load([config_path])
        else:
            print('Global config was not created.')

    def edit_config(self, type=None):
        """Open chosen configuration file in text editor."""
        editor = config['GENERAL'].get('editor')
        if type == 'local' or type is None:
            from .config import local_config as path
        elif type == 'global':
            from .config import user_config as path
        elif type.isdigit() is True:
            id = self._check_id(type)
            path = os.path.join(self.root, f'jobs/{id}/pydin.ini')
        path = os.path.abspath(path)
        if os.path.exists(path) is True:
            # Launch edition process and wait until it is completed.
            proc = sp.Popen([editor, path])
            print(f'Editing {path}...')
            proc.wait()
        else:
            print(f'File {path} does not exists.')

    def create_repo(self):
        """Create job repository and publish it by URL."""
        url = input('Please give remote Git repository URL: ')
        sure = self.sure or self._are_you_sure()
        if sure:
            self.driver.create_repo(url=url)
            print('Job repository successfully created.')

    def sync_repo(self, id=None):
        """Synchronize job repository."""
        print('You are trying to synchronize job repository...')
        sure = self.sure or self._are_you_sure()
        if sure:
            self.driver.pull_repo()
            kwargs = {}
            if id is not None:
                kwargs['id'] = int(id)
            self.driver.push_repo(**kwargs)
            print('Job repository successfully synchronized.')

    def _parse_arguments(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('-y', '--yes', dest='sure', action='store_true',
                            required=False, help='do not ask to confirm')
        parser.add_argument('-w', '--wait', dest='wait', action='store_true',
                            required=False, help='wait for job to finish')
        parser.add_argument('-o', '--output', dest='output',
                            action='store_true', required=False,
                            help='show process stdout')
        parser.add_argument('-e', '--error', dest='error',
                            action='store_true', required=False,
                            help='show process stderr')
        args, anons = parser.parse_known_args()
        return args

    def _header(self):
        """Print main application header."""
        path = os.path.join(os.path.dirname(__file__), 'samples/head.txt')
        text = open(path, 'r').read()
        print(text)

    def _are_you_sure(self):
        while True:
            sure = input('Are you sure? [Y/n] ')
            if sure in ('Y', 'n'):
                sure = True if sure == 'Y' else False
                break
        return sure

    def _configure_database(self):
        """Configure database connection."""
        from .config import user_config
        print('Configure the listed options to set up the DB schema.')
        print('Application DB schema will be automatically deployed.')
        vendor_name = input('Enter your DB vendor name: ')
        config['DATABASE']['vendor_name'] = vendor_name
        if vendor_name == 'sqlite':
            path = os.path.abspath('pydin.sqlite3')
            if os.path.exists(path):
                print(f'SQLite DB file {path} already exists!')
                return
            else:
                config['DATABASE']['path'] = path
                db.configure()
        elif vendor_name == 'oracle':
            print('Note: you can skip this items '
                  'and set them up later yourself.')
            driver_name = 'cx_oracle'
            config['DATABASE']['driver_name'] = driver_name
            host = input('DB host address: ')
            port = input('DB port number: ')
            sid = input('DB SID: ')
            service_name = input('DB service name: ')
            username = input('User name: ')
            password = getpass.getpass('User password: ')
            for k, v in dict(host=host, port=port, sid=sid,
                             service_name=service_name,
                             username=username, password=password).items():
                if v:
                    config['DATABASE'][k] = v
            if host and port:
                if sid or service_name:
                    if username and password:
                        db.configure()
            else:
                print(f'Please configure the DB connection in {user_config}. '
                      f'Then deploy the DB schema using scripts from GitHub.')
                return
        else:
            print(f'Sorry. This DB vendor {vendor_name} is not supported.')
            return
        config.save()
        print(f'Schema configured. Check configuration file {user_config}.')
        while True:
            sure = input('Continue? [Y/n] ')
            if sure in ('Y', 'n'):
                if sure == 'Y':
                    db.deploy()
                    if vendor_name == 'sqlite':
                        print(f'Schema deployed in SQLite DB file {path}.')
                    elif vendor_name == 'oracle':
                        address = f'{host}/{sid or service_name}'
                        print(f'Schema deployed in Oracle DB {address}.')
                break

    def _configure_job(self):
        name = input('Name this job:\n') or None
        desc = input('Describe shortly this job:\n') or None
        env = input('Enter the environment or leave empty to use default: ')
        env = env or None

        print()
        print('Schedule this job:')
        mday = input('Month day [1-31] ==== ') or None
        hour = input('Hour [0-23] ========= ') or None
        min = input('Minute [0-59] ======= ') or None
        sec = input('Second [0-59] ======= ') or None
        wday = input('Week day [1-7] ====== ') or None
        yday = input('Year day [1-366] ==== ') or None
        trig = input('Parent Job ID(s) [1...n] ') or None
        trig_list = None
        if isinstance(trig, str):
            if trig.isdigit() or trig == '-':
                trig = int(trig) if trig.isdigit() else trig
                trig = db.null if trig == '-' else trig
                trig_list = db.null
            elif ';' in trig:
                trig_list = trig
                trig = db.null
            else:
                trig = to_none(trig)
                trig_list = to_none(trig)

        print()
        start_date = input('Start date [YYYY-MM-DD HH24:MI:SS] ') or None
        start_date = db.null if start_date is None else None
        end_date = input('End date [YYYY-MM-DD HH24:MI:SS] ') or None
        end_date = db.null if end_date is None else None

        print()
        timeout = input('Set timeout [1...n] ')
        timeout = int(timeout) if timeout.isdigit() else None
        timeout = db.null if timeout is None else timeout
        parallelism = input('Set parallelism degree [1...n] ') or None
        if isinstance(parallelism, str):
            if parallelism.isdigit() or parallelism == 'N':
                parallelism = 'N' if parallelism == '1' else parallelism
            else:
                parallelism = to_none(parallelism)

        print()
        rerun_interval = input('Interval between reruns [1...n] ') or None
        if isinstance(rerun_interval, str):
            if rerun_interval.isdigit() or rerun_interval == '-':
                if rerun_interval.isdigit():
                    rerun_interval = int(rerun_interval)
                else:
                    rerun_interval = db.null
            else:
                rerun_interval = to_none(rerun_interval)
        rerun_limit = input('Limit of reruns number [1...n] ') or None
        if isinstance(rerun_limit, str):
            if rerun_limit.isdigit() or rerun_limit == '-':
                if rerun_limit.isdigit() and rerun_limit != '0':
                    rerun_limit = int(rerun_limit)
                else:
                    rerun_limit = db.null
            else:
                rerun_limit = to_none(rerun_limit)
        rerun_days = input('Limit of days for reruns [1...n] ') or None
        if isinstance(rerun_days, str):
            if rerun_days.isdigit() or rerun_days == '-':
                if rerun_days.isdigit():
                    rerun_days = int(rerun_days)
                else:
                    rerun_days = db.null
            else:
                rerun_days = to_none(rerun_days)
        rerun_period = input('Minutes of rerun window [0-59] ') or None
        if isinstance(rerun_period, str):
            if rerun_period == '-':
                rerun_period = db.null

        print()
        sleep_period = input('Hours of sleep window [0-23] ') or None
        if isinstance(sleep_period, str):
            if sleep_period == '-':
                sleep_period = db.null
        wake_up_period = input('Minutes of wake up window [0-59] ') or None
        if isinstance(wake_up_period, str):
            if wake_up_period == '-':
                wake_up_period = db.null

        print()
        alarm = input('Enable alarming for this job [Y] ')
        alarm = True if alarm == 'Y' else None
        email_list = input('List notification email addresses [a,b,...]: ')
        pattern = r'([a-zA-Z0-9._-]+@[a-zA-Z0-9._-]+\.[a-zA-Z0-9_-]+)'
        email_list = ','.join(re.findall(pattern, email_list)) or None
        print()

        result = {'name': name,
                  'desc': desc,
                  'env': env,
                  'mday': mday,
                  'hour': hour,
                  'min': min,
                  'sec': sec,
                  'wday': wday,
                  'yday': yday,
                  'trig': trig,
                  'trig_list': trig_list,
                  'start_date': start_date,
                  'end_date': end_date,
                  'timeout': timeout,
                  'parallelism': parallelism,
                  'rerun_interval': rerun_interval,
                  'rerun_limit': rerun_limit,
                  'rerun_days': rerun_days,
                  'rerun_period': rerun_period,
                  'sleep_period': sleep_period,
                  'wake_up_period': wake_up_period,
                  'alarm': alarm,
                  'email_list': email_list}
        return result

    def _expand_configs(self):
        if not os.path.exists(connector.config):
            open(connector.config, 'w')

    def _check_id(self, id):
        if id.isdigit() is True:
            return int(id)
        else:
            raise TypeError('id must be digit')



"""Contains application constants."""

import platform


LINUX = True if platform.system() == 'Linux' else False
MACOS = True if platform.system() == 'Darwin' else False
WINDOWS = True if platform.system() == 'Windows' else False



"""Setup."""

import pydin
import setuptools


name = pydin.__name__
version = pydin.__version__
author = pydin.__author__
author_email = pydin.__email__
description = pydin.__doc__.splitlines()[0]
long_description = open('README.md', 'r').read()
long_description_content_type = 'text/markdown'
license = pydin.__license__
url = 'https://github.com/t3eHawk/pydin'
python_requires = '>=3.7'
install_requires = open('requirements.txt').read().splitlines()
packages = setuptools.find_packages()
package_data = {'pydin': ['demo/*', 'samples/*']}
classifiers = ['Programming Language :: Python :: 3.7',
               'Programming Language :: Python :: 3.11',
               'License :: OSI Approved :: MIT License',
               'Operating System :: OS Independent']


setuptools.setup(name=name,
                 version=version,
                 author=author,
                 author_email=author_email,
                 description=description,
                 long_description=long_description,
                 long_description_content_type=long_description_content_type,
                 license=license,
                 url=url,
                 python_requires=python_requires,
                 install_requires=install_requires,
                 packages=packages,
                 package_data=package_data,
                 classifiers=classifiers)



"""Contains decorators used in application."""

import git
import functools


class BaseProperty():
    def __init__(self, func):
        self.func = func

    def __set__(self, obj, value):
        class_name = obj.__class__.__name__
        field_name = self.func.__name__
        full_name = f'property {field_name} of {class_name}'
        message = f'{full_name} cannot be changed this way'
        raise AttributeError(message)


class ClassProperty(BaseProperty):
    def __get__(self, obj, cls):
        return self.func(cls)


class StaticProperty(BaseProperty):
    def __get__(self, obj, cls):
        return self.func()


def classproperty(func):
    """Class property implementation."""
    return ClassProperty(func)


def staticproperty(func):
    """Static property implementation."""
    return StaticProperty(func)


def check_repo(func):
    """Check if git repo exists and prevent access to API methods if not."""
    @functools.wraps(func)
    def wrapper(obj, *args, **kwargs):
        try:
            git.Repo(obj.jobs)
        except git.exc.InvalidGitRepositoryError:
            raise Exception(f'no git repo in {obj.jobs} found')
        else:
            func(obj, *args, **kwargs)
    return wrapper



"""Contains main application logger."""

import pepperoni as pe

from .config import config


logger = pe.logger('pydin.core.logger',
                   format='{isodate}  {thread}  {rectype}  {message}\n',
                   console=config['LOGGING'].get('console'),
                   file=config['LOGGING'].get('file'),
                   info=config['LOGGING'].get('info'),
                   debug=config['LOGGING'].get('debug'),
                   error=config['LOGGING'].get('error'),
                   warning=config['LOGGING'].get('warning'),
                   critical=config['LOGGING'].get('critical'),
                   alarming=config['LOGGING'].get('alarming'),
                   maxsize=config['LOGGING'].get('maxsize'),
                   maxdays=config['LOGGING'].get('maxdays'),
                   email=config['EMAIL'].get('toggle'),
                   smtp={'host': config['EMAIL'].get('host'),
                         'port': config['EMAIL'].get('port'),
                         'tls': config['EMAIL'].get('tls'),
                         'address': config['EMAIL'].get('address'),
                         'user': config['EMAIL'].get('username'),
                         'password': config['EMAIL'].get('password'),
                         'recipients': config['GENERAL'].get('owner')})
